/**
 * Cecilia, an implementation of the Fractal component model in C.
 *
 * Copyright (C) 2007 ST Microelectronics, INRIA
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Contact: fractal@objectweb.org
 *
 * Authors: Matthieu Leclercq, Lionel Debroux
 */

// START SNIPPET: Content
#include <stdio.h>
#include <string.h>

/** Declare component internal data */
DECLARE_DATA {
    // Quick & dirty implementation for now.
    // To make things more generic, an associative array should be used.
    // Two 3-clause-BSD-licensed (GPL-compatible and therefore LGPL-compatible)
    // implementations can be found through http://en.wikipedia.org/wiki/Associative_array#C
    // * http://www.cl.cam.ac.uk/~cwc22/hashtable/
    // * http://uthash.sourceforge.net/

    // We hold references to components implementing the comanche.RequestHandler interface.
    // The Cecilia IDL to C name mapping convention requires that "comanche.RequestHandler"
    // is turned into "Rcomanche_RequestHandler", which is a valid C identifier, unlike
    // "comanche.RequestHandler".
    /** Reference to a request handler (exact binding defined in a Cecilia ADL file) */
    Rcomanche_RequestHandler * h0;
    /** Reference to a request handler (exact binding defined in a Cecilia ADL file) */
    Rcomanche_RequestHandler * h1;
} ;

/** Include cecilia.h. Must be included after the DECLARE_DATA */
#include <cecilia.h>

#include "ErrorCodes.h"


// -----------------------------------------------------------------------------
// Implementation of the RequestHandler interface.
// -----------------------------------------------------------------------------
/**
 * This method is called by the request analyser.
 * Cecilia IDL declaration: int handleRequest(any r);
 */
int METHOD(rh, handleRequest) (void *_this, void * r) {  
    int result;
#ifdef DEBUG
    fprintf(stdout,"RequestDispatcher: handleRequest\n");
#endif

    // Quick & dirty implementation for now.
    if (DATA.h0 == NULL) {
        return Comanche_Tried_calling_NULL; // Error, not supposed to be NULL.
    }
    result = CALL(DATA.h0, handleRequest, r);
    // If the first handler could not handle the first request,
    // check why, and call the second handler if necessary.
    if (result == Comanche_Nonexistent_file) {
        if (DATA.h1 == NULL) {
            return Comanche_Tried_calling_NULL; // Error, not supposed to be NULL.
        }
        result = CALL(DATA.h1, handleRequest, r);
    }
    else {
        // Otherwise, either it succeeded, or an error occurred.
        // In both cases, bail out.
    }

    return result;
}


// -----------------------------------------------------------------------------
// Implementation of the BindingController interface. This interface can be
// reached through predefined name "binding_controller".
//
// The BindingController interface must be implemented by the request dispatcher
// because of the "collection" cardinality of the dependency to handlers.
// -----------------------------------------------------------------------------

/**
 * Implements BindingController listFc method (implementation based on
 * an implementation of listFc found in the library).
 * Cecilia IDL declaration: int listFc(const in string[] clientItfNames);
 */
int METHOD(binding_controller, listFc)(void* _this, const char* clientItfNames[]) {
    (void)_this;
    unsigned int i;
    // Quick & dirty implementation for now.
    static const char * const itfNames[] = {"h0", "h1"};

    if (clientItfNames != NULL) {
        for (i = 0; i < (sizeof(itfNames)/sizeof(itfNames[0])); i++) {
            clientItfNames[i] = (char *)itfNames[i];
        }
    }

    return (sizeof(itfNames)/sizeof(itfNames[0]));
}

/**
 * Implements BindingController lookupFc method.
 * Cecilia IDL declaration: int lookupFc(const in string clientItfName, out any interfaceReference);
 */
int METHOD(binding_controller, lookupFc)(void* _this, const char* clientItfName, void** interfaceReference) {
    // Quick & dirty implementation for now.
    if (!(strcmp(clientItfName,"h0"))) {
        *interfaceReference = (void *)(DATA.h0);
    }
    else if (!(strcmp(clientItfName,"h1"))) {
        *interfaceReference = (void *)(DATA.h1);
    }
    else {
        return fractal_api_ErrorConst_NO_SUCH_INTERFACE;
    }
    return fractal_api_ErrorConst_OK;
}

/**
 * Implements BindingController bindFc method.
 * Cecilia IDL declaration: int bindFc(const in string clientItfName, in any serverItf);
 */
int METHOD(binding_controller, bindFc)(void* _this, const char* clientItfName, void* serverItf) {
    // Quick & dirty implementation for now.
    if (!(strcmp(clientItfName,"h0"))) {
        DATA.h0 = (Rcomanche_RequestHandler *)serverItf;
    }
    else if (!(strcmp(clientItfName,"h1"))) {
        DATA.h1 = (Rcomanche_RequestHandler *)serverItf;
    }
    else {
        return fractal_api_ErrorConst_NO_SUCH_INTERFACE;
    }
    return fractal_api_ErrorConst_OK;
}

/**
 * Implements BindingController unbindFc method.
 * Cecilia IDL declaration: int unbindFc(const in string clientItfName);
 */
int METHOD(binding_controller, unbindFc)(void* _this, const char* clientItfName) {
    if (!(strcmp(clientItfName,"h0"))) {
        DATA.h0 = (Rcomanche_RequestHandler *)NULL;
    }
    else if (!(strcmp(clientItfName,"h1"))) {
        DATA.h1 = (Rcomanche_RequestHandler *)NULL;
    }
    else {
        return fractal_api_ErrorConst_NO_SUCH_INTERFACE;
    }
    return fractal_api_ErrorConst_OK;
}
